//-----------------------------------------------------------------------------
// Copyright (c) 2008-2011 dhpoware. All Rights Reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
//-----------------------------------------------------------------------------

using System;
using System.Text;
using Dhpoware;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;

namespace XNATerrainTexturing
{
    /// <summary>
    /// This Windows XNA application demonstrates procedural terrain texturing
    /// using 4 tileable terrain textures and a height map. The terrain is
    /// randomly generated using a diamond-square (midpoint displacement)
    /// algorithm. 4 terrain region types are defined based on the heights of
    /// the generated terrain's height map. The terrain is painted using a
    /// custom effect shader.
    /// </summary>
    public class Demo : Microsoft.Xna.Framework.Game
    {
        private static void Main()
        {
            using (Demo demo = new Demo())
            {
                demo.Run();
            }
        }

        private GraphicsDeviceManager graphics;
        private SpriteBatch spriteBatch;
        private SpriteFont spriteFont;
        private Vector2 fontPos;
        private KeyboardState currentKeyboardState;
        private KeyboardState prevKeyboardState;
        private int windowWidth;
        private int windowHeight;
        private int frames;
        private int framesPerSecond;
        private TimeSpan elapsedTime = TimeSpan.Zero;
        private bool displayHelp;
        private CameraComponent camera;
        private Terrain terrain;

        public Demo()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";

            Window.Title = "XNA 4.0 Terrain Texturing";
            IsMouseVisible = false;
            IsFixedTimeStep = false;

            camera = new CameraComponent(this);
            Components.Add(camera);
        }     

        protected override void Initialize()
        {
            // Setup the window to be a quarter the size of the desktop.
            windowWidth = GraphicsDevice.DisplayMode.Width / 2;
            windowHeight = GraphicsDevice.DisplayMode.Height / 2;

            // Setup frame buffer.
            graphics.SynchronizeWithVerticalRetrace = false;
            graphics.PreferredBackBufferWidth = windowWidth;
            graphics.PreferredBackBufferHeight = windowHeight;
            graphics.PreferMultiSampling = true;
            graphics.ApplyChanges();

            // Position the text.
            fontPos = new Vector2(1.0f, 1.0f);

            // Setup the initial input states.
            currentKeyboardState = Keyboard.GetState();

            base.Initialize();
        }

        private bool KeyJustPressed(Keys key)
        {
            return currentKeyboardState.IsKeyDown(key) && prevKeyboardState.IsKeyUp(key);
        }

        protected override void LoadContent()
        {
            spriteBatch = new SpriteBatch(GraphicsDevice);
            spriteFont = Content.Load<SpriteFont>(@"Fonts\DemoFont");

            // Setup terrain.

            terrain = new Terrain(GraphicsDevice, Content);
            terrain.Create(256, 16, 0.0f, 1024.0f);

            terrain.AddRegion(  0.0f,  300.0f, @"Textures\region1");
            terrain.AddRegion(301.0f,  501.0f, @"Textures\region2");
            terrain.AddRegion(502.0f,  802.0f, @"Textures\region3");
            terrain.AddRegion(803.0f, 1024.0f, @"Textures\region4");

            terrain.SunlightDirection = Vector3.Down;
            terrain.SunlightColor = new Vector4(1.0f, 1.0f, 1.0f, 1.0f);
            terrain.TerrainAmbient = new Vector4(0.3f, 0.3f, 0.3f, 1.0f);
            terrain.TerrainDiffuse = new Vector4(0.7f, 0.7f, 0.7f, 1.0f);

            terrain.GenerateUsingDiamondSquareFractal(1.2f);

            // Setup camera.

            camera.CurrentBehavior = Camera.Behavior.Spectator;
            camera.Velocity = new Vector3(300.0f, 300.0f, 300.0f);
            camera.Acceleration = new Vector3(600.0f, 600.0f, 600.0f);
            
            camera.Perspective(85.0f, (float)windowWidth / (float)windowHeight,
                1.0f, terrain.TerrainBoundingSphere.Radius * 4.0f);
            
            Vector3 cameraPos = new Vector3();
            cameraPos.X = terrain.TerrainSize / 2.0f;
            cameraPos.Z = terrain.TerrainSize * 0.75f;
            cameraPos.Y = terrain.TerrainHeightMap.HeightAt(cameraPos.X, cameraPos.Z) * 1.5f;

            camera.LookAt(cameraPos, terrain.TerrainCenter, Vector3.Up);
        }

        private void ProcessKeyboard()
        {
            prevKeyboardState = currentKeyboardState;
            currentKeyboardState = Keyboard.GetState();

            if (KeyJustPressed(Keys.Escape))
                this.Exit();

            if (KeyJustPressed(Keys.H))
                displayHelp = !displayHelp;

            if (currentKeyboardState.IsKeyDown(Keys.LeftAlt) ||
                currentKeyboardState.IsKeyDown(Keys.RightAlt))
            {
                if (KeyJustPressed(Keys.Enter))
                    ToggleFullScreen();
            }

            if (KeyJustPressed(Keys.Space))
                terrain.GenerateUsingDiamondSquareFractal(1.2f);

            if (KeyJustPressed(Keys.D1))
            {
                camera.CurrentBehavior = Camera.Behavior.FirstPerson;
                camera.Velocity = new Vector3(200.0f, 200.0f, 200.0f);
                camera.Acceleration = new Vector3(400.0f, 400.0f, 400.0f);
            }

            if (KeyJustPressed(Keys.D2))
            {
                camera.CurrentBehavior = Camera.Behavior.Spectator;
                camera.Velocity = new Vector3(300.0f, 300.0f, 300.0f);
                camera.Acceleration = new Vector3(600.0f, 600.0f, 600.0f);
            }

            if (KeyJustPressed(Keys.D3))
            {
                camera.CurrentBehavior = Camera.Behavior.Flight;
                camera.Velocity = new Vector3(300.0f, 300.0f, 300.0f);
                camera.Acceleration = new Vector3(600.0f, 600.0f, 600.0f);
            }
        }

        private void ToggleFullScreen()
        {
            int newWidth = 0;
            int newHeight = 0;

            graphics.IsFullScreen = !graphics.IsFullScreen;

            if (graphics.IsFullScreen)
            {
                newWidth = GraphicsDevice.DisplayMode.Width;
                newHeight = GraphicsDevice.DisplayMode.Height;
            }
            else
            {
                newWidth = windowWidth;
                newHeight = windowHeight;
            }

            graphics.PreferredBackBufferWidth = newWidth;
            graphics.PreferredBackBufferHeight = newHeight;
            graphics.ApplyChanges();
        }

        protected override void Update(GameTime gameTime)
        {
            if (!IsActive)
                return;

            base.Update(gameTime);

            ProcessKeyboard();
            UpdateCamera(gameTime);
            UpdateFrameRate(gameTime);
        }

        private void UpdateCamera(GameTime gameTime)
        {
            Vector3 pos = camera.Position;
            float gridSpacing = terrain.TerrainHeightMap.GridSpacing;
            float size = terrain.TerrainHeightMap.Size * gridSpacing;
            float lowerBounds = 2.0f * gridSpacing;
            float upperBounds = size - (2.0f * gridSpacing);
            float height = terrain.TerrainHeightMap.HeightAt(pos.X, pos.Z) + 25.0f;

            if (pos.X < lowerBounds)
                pos.X = lowerBounds;

            if (pos.X > upperBounds)
                pos.X = upperBounds;
            
            switch (camera.CurrentBehavior)
            {
            case Camera.Behavior.FirstPerson:
                pos.Y = height;
                break;

            case Camera.Behavior.Spectator:
            case Camera.Behavior.Flight:
                if (pos.Y < height)
                    pos.Y = height;

                if (pos.Y > terrain.TerrainSize * 0.5f)
                    pos.Y = terrain.TerrainSize * 0.5f;
                break;

            default:
                break;
            }
            
            if (pos.Z < lowerBounds)
                pos.Z = lowerBounds;

            if (pos.Z > upperBounds)
                pos.Z = upperBounds;

            camera.Position = pos;
        }

        private void UpdateFrameRate(GameTime gameTime)
        {
            elapsedTime += gameTime.ElapsedGameTime;

            if (elapsedTime > TimeSpan.FromSeconds(1))
            {
                elapsedTime -= TimeSpan.FromSeconds(1);
                framesPerSecond = frames;
                frames = 0;
            }
        }

        private void IncrementFrameCounter()
        {
            ++frames;
        }
                
        private void DrawText()
        {
            StringBuilder buffer = new StringBuilder();

            if (displayHelp)
            {
                buffer.AppendLine("Press 1 switch to first person camera mode");
                buffer.AppendLine("Press 2 switch to spectator camera mode");
                buffer.AppendLine("Press 3 switch to flight camera mode");
                buffer.AppendLine();
                buffer.AppendLine("Press SPACE to generate a new terrain");
                buffer.AppendLine("Press ALT and ENTER to toggle full screen");
                buffer.AppendLine("Press ESCAPE to exit");
                buffer.AppendLine();
                buffer.AppendLine("Press H to hide help");
            }
            else
            {
                string cameraMode = null;

                switch (camera.CurrentBehavior)
                {
                case Camera.Behavior.FirstPerson:
                    cameraMode = "First person";
                    break;

                case Camera.Behavior.Spectator:
                    cameraMode = "Spectator";
                    break;

                case Camera.Behavior.Flight:
                    cameraMode = "Flight";
                    break;

                default:
                    break;
                }

                buffer.AppendFormat("FPS: {0}\n", framesPerSecond);
                buffer.AppendFormat("Camera mode: {0}\n", cameraMode);
                buffer.AppendLine();
                buffer.AppendLine("Press H to display help");
            }

            spriteBatch.Begin(SpriteSortMode.Deferred, BlendState.AlphaBlend);
            spriteBatch.DrawString(spriteFont, buffer.ToString(), fontPos, Color.Yellow);
            spriteBatch.End();
        }

        protected override void Draw(GameTime gameTime)
        {
            if (!IsActive)
                return;

            GraphicsDevice.Clear(Color.CornflowerBlue);

            GraphicsDevice.BlendState = BlendState.Opaque;
            GraphicsDevice.DepthStencilState = DepthStencilState.Default;
            GraphicsDevice.SamplerStates[0] = SamplerState.LinearWrap;
            GraphicsDevice.SamplerStates[1] = SamplerState.LinearWrap;
            GraphicsDevice.SamplerStates[2] = SamplerState.LinearWrap;
            GraphicsDevice.SamplerStates[3] = SamplerState.LinearWrap;

            terrain.Draw(Matrix.Identity, camera.ViewMatrix, camera.ProjectionMatrix);
            DrawText();

            base.Draw(gameTime);
            IncrementFrameCounter();
        }
    }
}